/*
   ftok.c
   Written by Richard Dawe <richdawe@bigfoot.com>
   
   Implementation of ftok(), according to Unix98 spec.
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include <sys/ipc.h>

/* Internal structures for keeping track of ftok() calls */
typedef struct _FTOK_INFO {
	const char *path;
	int         id;
	key_t       key;
} FTOK_INFO;

/* Limit maximum number of shared memory keys */
#define FTOK_INFO_MAX 256
static FTOK_INFO ftok_info[FTOK_INFO_MAX];
static int ftok_seed;
static int ftok_next_key;

/* -------------
 * - ftok_init -
 * ------------- */

static int ftok_init (void)
{
	static int ftok_inited = 0;
	int i;

	if (ftok_inited) return(ftok_inited);

	/* Blank the memory */
	for (i = 0; i < FTOK_INFO_MAX; i++) {
		bzero(&ftok_info[i], sizeof(FTOK_INFO_MAX));
	}

	/* Set up the key numbering */
	ftok_next_key = ftok_seed = rand() % FTOK_INFO_MAX;

	ftok_inited = 1;
	return(1);
}

/* ftok_init() should be called automatically. */
void __attribute__((constructor)) ftok_constructor (void) { ftok_init(); }

/* --------
 * - ftok -
 * -------- */

key_t ftok (const char *path, int id)
{
	struct stat sbuf;
	int my_id = id & 0xff; /* Only id's lower 8 bits are significant */
	int i;

	/* Initialise the buffers */
	if (!ftok_init()) {
		errno = ENOMEM;
		return(-1);
	}

	/* Behaviour unspec'd if id == 0, so let's be nice and return an
	 * error. */
	if (id == 0) {
		errno = EINVAL;
		return(-1);
	}

	/* Path must be stat'able - propagate stat's errno */
	if (stat(path, &sbuf) != 0) return(-1);

	/* Find any previous references to this path-id combo */
	for (i = 0; i < FTOK_INFO_MAX; i++) {
		if (   (my_id == ftok_info[i].id)
		    && (strcmp(path, ftok_info[i].path) == 0) ) {
			return(ftok_info[i].key);
		}
	}

	/* Not here => new key */
	for (i = 0; i < FTOK_INFO_MAX; i++) {
		if (ftok_info[i].id != 0) continue;

		/* Store the key details */
		ftok_info[i].id   = my_id;
		ftok_info[i].path = path;
		ftok_info[i].key  = ftok_next_key;

		/* Calculate next key */
		ftok_next_key++, ftok_next_key %= FTOK_INFO_MAX;
		return(ftok_info[i].key);
	}

	/* Unable to complete call */
	errno = ENOMEM;
	return(-1);
}
