/*
 * posixshm.c - POSIX-y shared memory functions
 * Written by Richard Dawe <richdawe@bigfoot.com>
 *
 * This makes use of Christopher Nelson's /dev/ramdrv DJGPP File System
 * Extension - see the files 'mmapio.c' & 'mmio.h' for more details.
 */

/*
 * These shared memory functions are really from the Unix98 spec, but they
 * appear to be derived from the POSIX Realtime Extension, hence the name
 * posixshm. I suppose it should be posixyshm, but that doesn't fit into
 * 8 chars :(
 *
 * The shared memory created by these functions does not share its namespace
 * with those in shm.c - i.e. shm created using ftok(), etc. on a path will
 * not conflict with that created with shm_open() on the same path.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <io.h>
#include <sys/fsext.h>

#include "mmio.h"
#include "posixshm.h"

/* This structure is used to keep track of the shared memory segment. There is
 * one per shared memory segment. */

typedef struct _posix_shm {
	int           id;          /* id is between 0 & 0xFFFF inclusive  */
	int           count;       /* Reference count                     */
	char         *name;        /* Path used in shm_open()             */
	char         *mmapio_name; /* Path passed to open() to create shm */
	int           mmapio_fd;   /* fd for accessing shm                */

	/* Opening flags - these cannot be overriden by shm_open() calls,
	 * because the creator is the owner. */
	int           oflag;
	mode_t        mode;
	uid_t         uid;
	gid_t         gid;
	uid_t         euid;
	gid_t         egid;

	/* If shm_unlink() has been called on this shm, then it should
	 * disappear from the name space. The shm segment will exist until
	 * it has been closed by all processes using it. In this way one
	 * could shm_unlink() some shm and then create some more with the
	 * sam path, but they would not overlap. */
	unsigned int  deleted:1;
} POSIX_SHM_INFO;

/* This structure keeps track of each shm file descriptor. It refers to a
 * shm segment descriptor (POSIX_SHM_INFO structure). */

typedef struct _posix_shm_fsext_data {
	POSIX_SHM_INFO *info;  /* Pointer to shm's info struct     */
	off_t           pos;   /* Current offset into mmapio file  */

	/* Open flags used by this instance */
	int             oflag;
	mode_t          mode;
	uid_t           uid;
	gid_t           gid;
	uid_t           euid;
	gid_t           egid;
} POSIX_SHM_FSEXT_DATA;

#define POSIX_SHM_NAME_PREFIX "/dev/ramdrv/"

#define POSIX_SHM_INFO_MAX 256
static POSIX_SHM_INFO posix_shm_info[POSIX_SHM_INFO_MAX];

/* ------------------
 * - posix_shm_init -
 * ------------------ */

static int posix_shm_init (void)
{
	static int posix_shm_inited = 0;
	int i;

	if (posix_shm_inited) return(posix_shm_inited);

	/* Clear info structs */
	for (i = 0; i < POSIX_SHM_INFO_MAX; i++) {
		bzero(&posix_shm_info[i], sizeof(POSIX_SHM_INFO));
	}

	/* Set up the memory-mapped I/O handler */
	mmf_constructor();

	posix_shm_inited = 1;
	return(posix_shm_inited);
}

/* posix_shm_init() should be called automatically. */
void __attribute__((constructor)) posix_shm_constructor (void) {
	posix_shm_init();
}

/* TODO: atexit() handler? */

/* -------------------
 * - posix_shm_fsext -
 * ------------------- */

static int posix_shm_fsext (__FSEXT_Fnumber func_number, int *rv, va_list args)
{
	__FSEXT_Function *f = NULL;
	POSIX_SHM_FSEXT_DATA *psfdata = NULL;
	POSIX_SHM_INFO *psinfo = NULL;
	int fd;
	void *buf = NULL;
	size_t size;
	off_t offset;
	int whence;
	struct stat *ss = NULL;

	switch (func_number) {
		/* No operation */
		case __FSEXT_nop:
			*rv = 0;
			break;

		/* Read */
		case __FSEXT_read:
			fd   = va_arg(args, int);
			buf  = va_arg(args, void *);
			size = va_arg(args, size_t);

			psfdata = __FSEXT_get_data(fd);
			if (psfdata == NULL) {
				errno = EBADF;
				*rv = -1;
				break;
			}

			/* Since there may be multiple fds referencing this
			 * shm segment, we have to seek to our previous
			 * position every time we want to read. */
			lseek(psfdata->info->mmapio_fd, SEEK_SET, psfdata->pos);
			*rv = read(psfdata->info->mmapio_fd, buf, size);
			if (*rv >= 0) psfdata->pos += *rv;
			break;
			
		/* Write */
		case __FSEXT_write:
			fd   = va_arg(args, int);
			buf  = va_arg(args, void *);
			size = va_arg(args, size_t);

			psfdata = __FSEXT_get_data(fd);
			if (psfdata == NULL) {
				errno = EBADF;
				*rv = -1;
				break;
			}

			/* Since there may be multiple fds referencing this
			 * shm segment, we have to seek to our previous
			 * position every time we want to write. */
			lseek(psfdata->info->mmapio_fd, SEEK_SET, psfdata->pos);
			*rv = write(psfdata->info->mmapio_fd, buf, size);
			if (*rv >= 0) psfdata->pos += *rv;
			break;

		/* Select */
		case __FSEXT_ready:
			fd = va_arg(args, int);

			psfdata = __FSEXT_get_data(fd);
			if (psfdata == NULL) {
				errno = EBADF;
				*rv = -1;
				break;
			}

			f = __FSEXT_get_function(psfdata->info->mmapio_fd);
			if (f == NULL) {
				errno = ENOSYS;
				*rv = -1;
				break;
			}

			/* TODO: Dodgy? This avoids calling select(), but is
			 * it a valid way of doing this? I suppose we are
			 * just reflecting the ready-check to the right
			 * FSEXT. */
			if (   f(__FSEXT_ready, rv, &psfdata->info->mmapio_fd)
			    != 1) {
				/* TODO: Failure? */
			}
			/* rv should be set correctly by inter-FSEXT call. */
			break;

		/* Close */
		case __FSEXT_close:
			fd = va_arg(args, int);

			psfdata = __FSEXT_get_data(fd);
			if (psfdata == NULL) {
				errno = EBADF;
				*rv = -1;
				break;
			}

			psinfo = psfdata->info;

			/* Close this reference. */
			__FSEXT_set_data(fd, NULL);
			free(psfdata);
			__FSEXT_set_function(fd, NULL);
			_close(fd);

			/* Decrement the reference counter. */
			psinfo->count--;

			/* If the shm segment has been deleted & the counter
			 * is now zero, remove it from memory. */
			if (psinfo->deleted && (psinfo->count == 0)) {
				close(psinfo->mmapio_fd);
				free(psinfo->name);
				free(psinfo->mmapio_name);
				bzero(psinfo, sizeof(POSIX_SHM_INFO));
			}

			*rv = 0;
			break;

		/* Seek */
		case __FSEXT_lseek:
			fd     = va_arg(args, int);
			offset = va_arg(args, off_t);
			whence = va_arg(args, int);

			psfdata = __FSEXT_get_data(fd);
			if (psfdata == NULL) {
				errno = EBADF;
				*rv = -1;
				break;
			}

			/* Seek the mmapio file and use the return from that. */
			*rv = lseek(psfdata->info->mmapio_fd, offset, whence);
			psfdata->pos = *rv;
			break;

		/* Open-file statistics */
		case __FSEXT_fstat:
			fd = va_arg(args, int);
			ss = va_arg(args, struct stat *);

			psfdata = __FSEXT_get_data(fd);
			if (psfdata == NULL) {
				errno = EBADF;
				*rv = -1;
				break;
			}

			*rv = fstat(psfdata->info->mmapio_fd, ss);
			break;

		/* Function not supported/implemented */
		case __FSEXT_open:   /* Caller should use shm_open()!  */
		case __FSEXT_creat:  /* Caller should use shm_open()!  */
		case __FSEXT_fcntl:  /* TODO: Write it! Support what?  */
		case __FSEXT_ioctl:  /* TODO: Support it?              */
		case __FSEXT_link:   /* Hmm                            */
		case __FSEXT_dup:    /* TODO: Support it?              */
		case __FSEXT_dup2:   /* TODO: Support it?              */

		/* TODO: The issue with these two subfunctions is whether the
		 * shared memory "file" names are part of the same namespace
		 * as the "normal" file system. Since shm_open() and
		 * shm_unlink() functions are provided, I suppose that shm
		 * and "real" files are in different namespaces. */
		case __FSEXT_unlink: /* TODO: Should use shm_unlink()? */
		case __FSEXT_stat:

		default:
			errno = ENOSYS;
			*rv = -1;
	}

	/* Don't let DOS get anywhere near this fd! */
	return(1);
}

/* ------------
 * - shm_open -
 * ------------ */

int shm_open (const char *name, int oflag, mode_t mode)
{
	char mmapio_name[PATH_MAX];
	POSIX_SHM_FSEXT_DATA *psfdata = NULL;
	int fd;
	int found, foundblank, shmidx;
	int i;

	/* Try to find a path match. */
	for (found = foundblank = -1, i = 0; i < POSIX_SHM_INFO_MAX; i++) {
		if (posix_shm_info[i].name == NULL) {
			/* Save blank occurrence to avoid rescanning later */
			if (foundblank == -1) foundblank = i;
			continue;
		}

		if (   (strcmp(posix_shm_info[i].name, name) == 0)
		    && !posix_shm_info[i].deleted) {
			found = i;
			break;
		}
	}

	/* Matched? */
	if (found != -1) {
		/* Yep - check with shm segment's info */

		/* Create & exclusive => failure, but create alone is OK. */
		if (   ((oflag & O_CREAT) == O_CREAT)
		    && ((oflag & O_EXCL)  == O_EXCL)) {
			errno = EEXIST;
			return(-1);
		}

		/* TODO: Check permissions */

		shmidx = found;
	} else {
		/* No space in array! */
		if (foundblank == -1) {
			errno = ENOSPC;
			return(-1);
		}

		/* Creation flags OK? */
		if ((oflag & O_CREAT) == 0) {
			errno = ENOENT;
			return(-1);
		}

		/* Create a mmapio file for the shm. */
		sprintf(mmapio_name, "%spshm%04x",
		        POSIX_SHM_NAME_PREFIX, shmidx);

		/* TODO: Make shm_open() options and open() options
		 * consistent. */
		fd = open(mmapio_name,
		          O_RDWR | O_CREAT | O_BINARY, /* Always create RDWR */
			  S_IRUSR | S_IWUSR);

		if (fd == -1) {
			errno = ENOSPC;
			return(-1);
		}

		/* Set up the info struct */
		shmidx = foundblank;

		bzero(&posix_shm_info[shmidx], sizeof(posix_shm_info[i]));
		posix_shm_info[shmidx].id          = shmidx;
		posix_shm_info[shmidx].name        = strdup(name);
		posix_shm_info[shmidx].mmapio_name = strdup(mmapio_name);
		posix_shm_info[shmidx].mmapio_fd   = fd;
		posix_shm_info[shmidx].oflag       =   (oflag & ~O_RDONLY)
		                                     | O_RDWR;
		posix_shm_info[shmidx].mode        = mode;
		posix_shm_info[shmidx].uid         = getuid();
		posix_shm_info[shmidx].gid         = getgid();
		posix_shm_info[shmidx].euid        = geteuid();
		posix_shm_info[shmidx].egid        = getegid();
	}

	/* TODO: Allow O_TRUNC flag to truncate mmapio file. */

/* If an operation fails for some reason, clean-up the info struct, if there
 * are no references to it, i.e. when the shm segment has only just been
 * created. */
#define __CLEAN_UP_PSI(i)						\
	if (posix_shm_info[i].count == 0) {				\
		free(posix_shm_info[i].name);				\
		free(posix_shm_info[i].mmapio_name);			\
		close(posix_shm_info[i].mmapio_fd);			\
		bzero(&posix_shm_info[i], sizeof(posix_shm_info[i]));	\
	}

	/* Get a new fd */
	fd = __FSEXT_alloc_fd(posix_shm_fsext);

	if (fd == -1) {
		/* Clean up */
		__CLEAN_UP_PSI(shmidx);

		errno = ENFILE; /* TODO: Error? */
		return(-1);
	}

	/* Allocate data struct for fd */
	psfdata = (POSIX_SHM_FSEXT_DATA *) malloc(sizeof(POSIX_SHM_FSEXT_DATA));
	if (psfdata == NULL) {
		/* Clean up */
		__FSEXT_set_function(fd, NULL);
		_close(fd);
		__CLEAN_UP_PSI(shmidx);

		errno = ENOSPC;
		return(-1);
	}

	bzero(psfdata, sizeof(POSIX_SHM_FSEXT_DATA));
	psfdata->info  = &posix_shm_info[shmidx];
	psfdata->pos   = 0;
	psfdata->oflag = oflag;
	psfdata->mode  = mode;
	psfdata->uid   = getuid();
	psfdata->gid   = getgid();
	psfdata->euid  = geteuid();
	psfdata->egid  = getegid();

	/* Get FSEXT to pass the data to our FSEXT */
	if (__FSEXT_set_data(fd, (void *) psfdata) == NULL) {
		/* Clean up */
		free(psfdata);
		__FSEXT_set_function(fd, NULL);
		_close(fd);
		__CLEAN_UP_PSI(shmidx);

		errno = ENOSPC; /* TODO: This seems appropriate, but? */
		return(-1);
	}

	/* Done */
	posix_shm_info[shmidx].count++;
	return(fd);
}

/* --------------
 * - shm_unlink -
 * -------------- */

int shm_unlink (const char *name)
{
	int found;
	int i;

	for (found = -1, i = 0; i < POSIX_SHM_INFO_MAX; i++) {
		if (   (strcmp(posix_shm_info[i].name, name) == 0)
		    && !posix_shm_info[i].deleted) {
			found = i;
			break;
		}
	}

	if (found == -1) {
		errno = ENOENT;
		return(-1);
	}

	/* TODO: Permissions? */

	/* No memory is freed until all references to the segment are
	 * destroyed, i.e. by close() on all refering file descriptors.
	 * Note that if shm_unlink() isn't called then the shm segment will
	 * exist indefinitely (well, until the threaded task terminates). */
	posix_shm_info[i].deleted = 1;

	/* Free the segment? */
	if (posix_shm_info[i].count == 0) {
		close(posix_shm_info[i].mmapio_fd);
		free(posix_shm_info[i].name);
		free(posix_shm_info[i].mmapio_name);
		bzero(&posix_shm_info[i], sizeof(POSIX_SHM_INFO));
	}

	return(0);
}
